
#include <stdlib.h>
#include <stdio.h>
#include "global.h"
#include "linesegment.h"
#include "polygon.h"


/**************************************************************************************************/

/* initialize polygon to empty list */
void 
polygon_init(Polygon* polygon)
{
    if( polygon == NULL ) return;

    polygon->head = NULL;
    polygon->size = 0;
}

/**************************************************************************************************/

/* 
 * destroys the list by freeing all memory allocated for each line segment returning the polygon
 * to an empty list
 */
void 
polygon_destroy(Polygon * polygon)
{
    if( polygon == NULL ) return;

    /* while the list of line segments is not empty, remove the element at the head of the list */
    while( polygon->size > 0 ) {
        polygon_remove_segment_next(polygon, NULL);
    }
}

/**************************************************************************************************/

/* 
 * inserts new line segment specified by (x1,y1), (x2, y2) into list after the given list element 
 * location. If the list element location is NULL, inserts line segment at head of list 
 */
bool 
polygon_insert_segment_next(Polygon *polygon, LineSegment *location, 
                                 float x1, float y1, float x2, float y2)
{
    LineSegment* new_segment;
    
    if( polygon == NULL ) return FALSE;

    new_segment = (LineSegment*)malloc(sizeof(LineSegment));
    if( new_segment == NULL ) return FALSE;

    /* initialize line segment coordinates to provided data values */
    new_segment->x1 = x1;
    new_segment->y1 = y1;
    new_segment->x2 = x2;
    new_segment->y2 = y2;
    
    /* insertion at head of list */
    if( location == NULL ) {
        new_segment->next = polygon->head;
        polygon->head = new_segment;
    }
    /* insertion somewhere other than head */
    else {
        new_segment->next = location->next;
        location->next = new_segment;
    }
    polygon->size++;
    return TRUE;
}

/**************************************************************************************************/

/* 
 * removes the line segment after the the given list element location. If the list element location
 * is NULL, removes the head of the list
 */
bool 
polygon_remove_segment_next(Polygon *polygon, LineSegment *location)
{
    LineSegment *old_segment;
    
    if( polygon == NULL || polygon->size == 0 ) return FALSE;
    
    if( location == NULL ) {
        old_segment = polygon->head;
        polygon->head = polygon->head->next;
    }
    else {
        if( location->next == NULL ) return FALSE;
        old_segment = location->next;
        location->next = location->next->next;
        
    }
    free(old_segment);
    polygon->size--;
    return TRUE;
}

/**************************************************************************************************/

/* returns a pointer to the line segment at the head of the list */
LineSegment *
polygon_head(Polygon *polygon) 
{
    if( polygon == NULL ) return NULL;
    return polygon->head;
}

/**************************************************************************************************/

/* returns the number of line segments within the polygon */
int 
polygon_num_segments(Polygon *polygon)
{
    if( polygon == NULL ) return -1;
    return polygon->size;
}

/**************************************************************************************************/

/* returns pointer to next line segment within the list */
LineSegment *
polygon_next_segment(LineSegment *segment) 
{
    if( segment == NULL ) return NULL;
    return segment->next;
}


/**************************************************************************************************/

/* 
 * searches the polygon to find a segment that start (or ends) at the end of the provided 
 * current segment,returning a pointer to the connected segment if it exists. Otherwise, returns NULL
 *
 * If the connected segment ends (x2, y2) at the end point of the current segment, the start and
 * end of that segment will be swapped. This ensures the algorithm within is_valid_polygon()
 * always follows from the start (x1, y1) to the end (x2, y2) of every connected line segment.
 */
LineSegment* find_next_connected_segment(Polygon* polygon, LineSegment* fromSegment) 
{
    LineSegment *toSegment;
    float tempVal;
    
    if( polygon == NULL ) return NULL;

    toSegment = polygon->head;
    
    /* while we have not reached the end of the list of line segments */
    while (toSegment != NULL ) {
        
        /* only conisder those segments that are not the current segment */
        if( toSegment != fromSegment ) {
            
            /* if end (x2, y2) of from segment is equal to start of the to segment */
            if(fromSegment->x2 == toSegment->x1 &&
               fromSegment->y2 == toSegment->y1 ) {
                
                /* return the found segment */
                return toSegment;
            }
            /* if end (x2, y2) of from segment is equal to end of the to segment */
            else if (fromSegment->x2 == toSegment->x2 &&
                     fromSegment->y2 == toSegment->y2 ) {
                
                /* sawp the start and end of the to line segment */
                tempVal = toSegment->x2;
                toSegment->x2 = toSegment->x1;
                toSegment->x1 = tempVal;
                tempVal = toSegment->y2;
                toSegment->y2 = toSegment->y1;
                toSegment->y1 = tempVal;
                
                /* return found segment */
                return toSegment;
            }
        }
        toSegment = polygon_next_segment(toSegment);
    }
    
    /* did not find a conneced segment */
    return NULL;
}

/**************************************************************************************************/

/* 
 * checks to determine if the specified polygon is a valid closed form polygon in which all segments 
 * are connected to another line segment
 *
 * NOTE: There is still a potential error with this implementation.
 */
bool is_valid_polygon(Polygon *polygon)
{
    LineSegment* segment;
    int segments_followed;
    
    /* start with first line segment in polygon */
    segment = polygon->head;
    segments_followed = 0;
    
    /* 
     * for each line segment found, search for another line segment that starts or ends
     * at the end point (x2, y2) of the current segment 
     *
     * continue as long as we have found another connected segment, we have not returned the
     * head of the list, and we have not followed more than the number of segments that are in
     * the polygon (i.e. make sure we don't get stuck in a cycle)
     */
    do {
        segment = find_next_connected_segment(polygon, segment);
        
        /* if the segment is not found, the polygon is invalid */
        if( segment == NULL ) return FALSE;
        
        /* if segment was found, increase the count of segments we have followed */
        segments_followed++;
        
    } while (segment != polygon->head && segments_followed != polygon->size);
    
    /* 
     * if we returned to the head in exactly the number of segments that exist in the 
     * polygon, we have a valid polygon shape 
     */
    if(segment == polygon->head && segments_followed == polygon->size) return TRUE;
    
    return FALSE;
}

/**************************************************************************************************/

/* prints list of line segments within polygon */
void 
polygon_display(Polygon *polygon) {
    LineSegment *segment;
    
    if( polygon == NULL ) return;
    
    segment = polygon->head;
    while ( segment != NULL ) {
        printf("(%f,%f)->(%f,%f)\n", segment->x1, segment->y1, segment->x2, segment->y2);
        segment = polygon_next_segment(segment);
    }
}

/**************************************************************************************************/

